"""Interval.  Like nilmdb.server.interval, but re-implemented here
in plain Python so clients have easier access to it.

Intervals are half-open, ie. they include data points with timestamps
[start, end)
"""

import nilmdb.utils.time
import nilmdb.utils.iterator

class IntervalError(Exception):
    """Error due to interval overlap, etc"""
    pass

# Interval
class Interval:
    """Represents an interval of time."""

    def __init__(self, start, end):
        """
        'start' and 'end' are arbitrary numbers that represent time
        """
        if start >= end:
            # Explicitly disallow zero-width intervals (since they're half-open)
            raise IntervalError("start %s must precede end %s" % (start, end))
        self.start = start
        self.end = end

    def __repr__(self):
        s = repr(self.start) + ", " + repr(self.end)
        return self.__class__.__name__ + "(" + s + ")"

    def __str__(self):
        return ("[" + nilmdb.utils.time.timestamp_to_string(self.start) +
                " -> " + nilmdb.utils.time.timestamp_to_string(self.end) + ")")

    def __cmp__(self, other):
        """Compare two intervals.  If non-equal, order by start then end"""
        return cmp(self.start, other.start) or cmp(self.end, other.end)

    def intersects(self, other):
        """Return True if two Interval objects intersect"""
        if not isinstance(other, Interval):
            raise TypeError("need an Interval")
        if self.end <= other.start or self.start >= other.end:
            return False
        return True

    def subset(self, start, end):
        """Return a new Interval that is a subset of this one"""
        # A subclass that tracks additional data might override this.
        if start < self.start or end > self.end:
            raise IntervalError("not a subset")
        return Interval(start, end)

def set_difference(a, b):
    """
    Compute the difference (a \\ b) between the intervals in 'a' and
    the intervals in 'b'; i.e., the ranges that are present in 'self'
    but not 'other'.

    'a' and 'b' must both be iterables.

    Returns a generator that yields each interval in turn.
    Output intervals are built as subsets of the intervals in the
    first argument (a).
    """
    # Iterate through all starts and ends in sorted order.  Add a
    # tag to the iterator so that we can figure out which one they
    # were, after sorting.
    def decorate(it, key_start, key_end):
        for i in it:
            yield i.start, key_start, i
            yield i.end, key_end, i
    a_iter = decorate(iter(a), 0, 2)
    b_iter = decorate(iter(b), 1, 3)

    # Now iterate over the timestamps of each start and end.
    # At each point, evaluate which type of end it is, to determine
    # how to build up the output intervals.
    a_interval = None
    b_interval = None
    out_start = None
    for (ts, k, i) in nilmdb.utils.iterator.imerge(a_iter, b_iter):
        if k == 0:
            # start a interval
            a_interval = i
            if b_interval is None:
                out_start = ts
        elif k == 1:
            # start b interval
            b_interval = i
            if out_start is not None and out_start != ts:
                yield a_interval.subset(out_start, ts)
            out_start = None
        elif k == 2:
            # end a interval
            if out_start is not None and out_start != ts:
                yield a_interval.subset(out_start, ts)
            out_start = None
            a_interval = None
        elif k == 3:
            # end b interval
            b_interval = None
            if a_interval:
                out_start = ts
